function [pA,b]=precondition(A,b,M1,flag,M2)
%[pA,pb]=precondition(A,b,M,flag)  
%  This function preconditions the matrix A and 
%  the right-hand side vector b by explicit preconditioning.
%
%  A defines the MV (as a matrix or as a function [by handle or by name]),
%  b is the right-hand side vector,
%  M defines the preconditioner (as a matrix, M\u, or as a function),
%  flag sets the type of preconditioning:
%    flag = 'r' for right preconditioning,
%           'l' for left preconditioning (default).
%
%  pA is the handle for the preconditioned matrix-vector multiplication,
%  pb is the preconditioned right-hand side vector.
%
%[pA,pb]=precondition(A,b,M1,flag,M2)
%  if flag = 'c' (central preconditioning), then M1 is the left preconditioner
%  and M2 is the right preconditioner. For other values of flag, the preconditioner
%  acts as the product M2*M1 of M1 and M2 (c=M2\M1\(A*u) in case flag='l').
%
%NOTE: also for left and right preconditioning, it can be more efficient to
%  use two matrices M1 and M2 rather than M=M1*M2 (MATLAB will solve M\u
%  by factorising M into M=L*U, which would be a waste of computational effort
%  if M1 and M2 are triangular).
%

  %%% define the MV
  c=lMV(u)
    if IsMatrix(A), c=A*u; else, c=feval(A,u); end
  end

  %%% define the (first) preconditioner
  c=lMsolve1(u)
    if IsMatrix(M1), c=M1\u; else, c=feval(M1,u); end
  end

  %%% define the second preconditioner
  c=lMsolve2(u)
    if nargin>4
      if IsMatrix(M2), c=M2\u; else, c=feval(M2,u); end
    else
      c=u;
    end
  end
 
  %%% find the type of preconditioning and precondition b
  switch lower(flag)
    case 'r'
      P=-1; 
    case 'c'
      P=0; b=lMsolve1(b);
    otherwise %%% preconditioning is left by default.
      P=1; b=lMsolve2(lMsolve1(b));
  end

  pA=@PreMV;

  %%% define the preconditioned matrix-vector muliplication
  c=PreMV(u)
    c=u;
    if P==1, c=lMV(c); c=lMsolve1(c); c=lMsolve2(c); end
    if P==0, c=lMsolve2(c); c=lMV(c); c=lMsolve1(c); end
    if P==-1,c=lMsolve1(c); c=lMsolve2(c); c=lMV(c); end
  end

end

