function [x,hist,t,NoMV]=cg(A,b,x0,tol,kmax,M1,M2)
%x=cg(A,b)
%  x is the computed solution (by [preconditioned] CG),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=cg(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=cg(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%
%x=cg(A,b,x0,tol,kmax,M)
%  M is the preconditioner, either a matrix, or a function-handle, or 
%    the name of a subroutine, say M='Msolve', that defines 
%    the action of the preconditioner, with u=Msolve(r) we have that u=M\r
%Note that the preconditioner "solves" rather than multiplies.
%
%x=cg(A,b,x0,tol,kmax,M1,M2)
%  M1 and M2 define the precondioner M as a product of two factors:
%    if M1 and M2 are matrices, then M=M1*M2, if M1='Msolve1' and M2='Msolve2'
%    are functions, then v=Msolve1(r) is v=M1\r and u=Msolve2(v) is u=M2\v;
%
%Note: if M=L*U, then matlab solves M*u=r as v=L\r; u=U\v;
%    however defining M=L*U rather than M1=L and M2=U requires Matlab to
%    factorize M into the L and U factors (does Matlab factorize in each step?)
%


  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end
  
%-------------------------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
  PRECOND=max(nargin-5,0);
  %%% if PRECOND>0, then a preconditioner has been defined
  %%% either as one precondioner or as a product of two factors

  if PRECOND, 
     M1_IS_MATRIX=IsMatrix(M1);
     if PRECOND>1, M2_IS_MATRIX=IsMatrix(M2); end
  end
  
  %%% define the preconditoner MSOLVE
  function u=MSOLVE(r)
    if M1_IS_MATRIX, u=M1\r; else, u=feval(M1,r); end
    if PRECOND==1, return, end
    if M2_IS_MATRIX, u=M2\u; else, u=feval(M2,u); end
  end
%-------------------------------------------------------------------------

  t0=clock; 
  %%%
  [n,m]=size(b); 
  u=zeros(n,1); rho1=1;

  x=x0; r=b(:,1); rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
      
  for k=1:kmax
    if rho<tol, break, end
    
    if PRECOND, c=MSOLVE(r); else, c=r; end

    rho0=rho1; rho1=r'*c; beta=-rho1/rho0;
    u=c-beta*u;
    
    c=MV(u);
    sigma=u'*c; alpha=rho1/sigma;
    x=x+alpha*u;
    r=r-alpha*c;
    %%% to save another dot product,
    %%% replace rho by rho1
    rho=norm(r); hist=[hist,rho];
  end
  t=etime(clock,t0);

end
