function [x,hist,t,NoMV]=minres(A,b,x0,tol,kmax)
%x=minres(A,b)
%  x is the computed solution (by MINRES),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=minres(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=minres(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%

  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end

%-------------------------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
%-------------------------------------------------------------------------
  t0=clock; 
  %%%
  [n,m]=size(b); 
   
  x=x0; r=b(:,1); rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
  
  %%% initialize Lanczos
  v=r/rho; v_old=zeros(n,1); beta=0;
  
  s=0; beta_t=0; c=-1;
  w=zeros(n,1); www=v;

  %%% scale tol for relative residual reduction
  tol=rho*tol;
  for k=1:kmax
    if abs(rho)<tol, break, end
    
    %%% Lanczos
    wv=MV(v)-beta*v_old;
    alpha=v'*wv; wv=wv-alpha*v;
    beta=norm(wv); v_old=v; v=wv/beta;
     
    %%% Update QR-decomposition of the Lanczos matrix
    l1=s*alpha - c*beta_t; l2=s*beta;
    alpha_t=-s*beta_t-c*alpha; beta_t=c*beta;
    l0 = sqrt(alpha_t*alpha_t+beta*beta); 
    c = alpha_t/l0; s = beta/l0;

    %%% update search vectors
    ww = www - l1*w; www = v - l2*w; w = ww/l0;
    %%% update approximate solution
    x =  x + (rho*c)*w; 
    %%% update residual norm
    rho =  s*rho; 
    
    hist=[hist,abs(rho)];
  end
  t=etime(clock,t0);

end
