function [x,hist,t,NoMV]=bicg(A,b,x0,tol,kmax,At)
%x=bicg(A,b)
%  x is the computed solution (by Bi-CG),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u,
%  if A is a matrix, then c=A'*u is used to multiply with the conjugate of A,
%  if A is function, then c=MV(u,'t') is used to multiply with the conjugate of A,
%  b is the right-hand side vector.
%
%x=bicg(A,b,x0,tol,kmax,At)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%  At is a matrix At, or a function-handle, say, A=@MVt, 
%    or a function-name, say A='MVt', that defines the action of A' on a vector: 
%    with c=MVt(u) we have that c=A'*u,
%
%[x,hist,t,NoMV]=bicg(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%


  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end

%---------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(u)
    if A_IS_MATRIX, c=A*u; else, c=feval(A,u); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
  TRANSPOSE=max(nargin-5,0);
  %%% TRANSPOSE > 1 if the transpose of A has explicitly been given
  if TRANSPOSE, %%% is the transpose a matrix?
     At_IS_MATRIX=IsMatrix(At);
  end
  
  function ct=MVt(ut)
   if TRANSPOSE
     if At_IS_MATRIX, ct=At*ut; else, ct=feval(At,ut); end 
   else %%% if the transpose of A is not given , use A
     if A_IS_MATRIX, ct=A'*ut; else, ct=feval(A,ut,'t'); end
   end
   NoMV=NoMV+1;
  end
%---------------------------------------------------------
 
  
  t0=clock; 
  %%%
  [n,m]=size(b); 
  u=zeros(n,1); mu=1; alpha=1;
  ut=zeros(n,1); 

  x=x0; r=b; rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
      
  rt=rand(n,1);  % rt=r; 
  k=0;
  while k<kmax
    if rho<tol, break, end
    
    sigma=-mu; 
    
    mu=rt'*r; beta=mu/sigma;
    u=r-beta*u;
    c=MV(u); k=k+1;
    sigma=rt'*c; alpha=mu/sigma;
    x=x+alpha*u;
    r=r-alpha*c;
    rho=norm(r);
    %%% history in terms of #MVs: one step requires two MVs (one by A and one by A') 
    hist=[hist,rho*ones(1,2)];
    
    ut=rt-beta*ut;
    ct=MVt(ut); k=k+1;
    rt=rt-alpha*ct;
    
  end
  t=etime(clock,t0);

end
