function [x,hist,t,NoMV]=bicgstabl(A,b,x0,tol,kmax,L)
%x=bicgstabl(A,b)
%  x is the computed solution (by BiCGstab(ell), ell=2 default),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=bicgstabl(A,b,x0,tol,kmax,ell)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%  ell, the degree of the factors of the stabilisation polynomials
%
%[x,hist,t,NoMV]=bicgstabl(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%

  %%% set defaults
  if nargin<6, L=2; end
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end
  
  warning off
  
%--------------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(u)
    if A_IS_MATRIX, c=A*u; else, c=feval(A,u); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
%-------------------------------------------------------------
  
  t0=clock; 
  %%%
  [n,m]=size(b); 
  
  u=zeros(n,L+1); sigma=1; omega=1;
  r=zeros(n,L+1); rt=rand(n,1);
  
  x=x0; r(:,1)=b(:,1); rho=norm(r(:,1));
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r(:,1)=r(:,1)-MV(x); rho=norm(r(:,1)); end
      hist=rho;
      
  k=0;
  while k<kmax
    if rho<tol, break, end
    
    sigma=-omega*sigma; 
    
    for l=1:L
      I=1:l;
      mu=rt'*r(:,l); beta=mu/sigma;
      u(:,I)=r(:,I)-beta*u(:,I);
      u(:,l+1)=MV(u(:,l)); k=k+1;
      
      sigma=rt'*u(:,l+1); alpha=mu/sigma;
      x=x+alpha*u(:,1);
      r(:,I)=r(:,I)-alpha*u(:,I+1);
      rho=norm(r(:,1)); hist=[hist,rho*ones(1,2)];
      if rho<tol, break, end
      r(:,l+1)=MV(r(:,l)); k=k+1;
    end

if 1
    G=r(:,I+1)'*r(:,I+1); gamma=G\(r(:,I+1)'*r(:,1));
    omega=gamma(l,1);
else
    [gamma,omega]=vanilla(r'*r);
end
                        
    x=x+r(:,I)*gamma;
    r(:,1)=r(:,1)-r(:,I+1)*gamma;
    u(:,1)=u(:,1)-u(:,I+1)*gamma;
    
    
    rho=norm(r(:,1)); hist(1,end)=rho;
  end
  t=etime(clock,t0);

end
%=============================
function [gamma,omega]=vanilla(G)
 
     kappa=0.7; %% kappa=0;
     l=size(G,2)-1; I=2:l; 
     Gamma=[-1,0;G(I,I)\G(I,[1,l+1]);0,-1];
     Ng=Gamma'*G*Gamma; 
     cosine=abs(Ng(1,2))/sqrt(Ng(1,1)*Ng(2,2)); omega=Ng(1,2)/Ng(2,2);

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %%% Increase |omega| for enhancing the stability of the BiCG part     %%%
  %%%   Sleijpen & Van der Vorst, Numerical Algor., 10 (1995) 203--223  %%%
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if cosine<kappa, omega=(kappa/cosine)*omega; end
    gamma=Gamma(2:l+1,1)-omega*Gamma(2:l+1,2); 

end
