function [x,hist,t,NoMV]=cgs(A,b,x0,tol,kmax,MVT)
%x=cgs(A,b)
%  x is the computed solution (by CGS),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=cgs(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=cgs(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%


  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end

%--------------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(u)
    if A_IS_MATRIX, c=A*u; else, c=feval(A,u); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
%-------------------------------------------------------------

  t0=clock; 
  %%%
  [n,m]=size(b); 
  u=zeros(n,1); w=zeros(n,1);
  sigma=1; alpha=1;


  x=x0; r=b(:,1); rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
      
  k=0; rt=rand(n,1);
  while k<kmax
    if rho<tol, break, end
       
    mu=rt'*r; beta=-(mu/sigma)/alpha;
    w=u-beta*w;
    u=r-beta*u;
    w=u-beta*w;
    c=MV(w); k=k+1;
    sigma=rt'*c; alpha=mu/sigma;
    u1=u-alpha*c;
    y=alpha*(u1+u); u=u1;
    x=x+y;
    c=MV(y); k=k+1;
    r=r-c;
    rho=norm(r); hist=[hist,rho*ones(1,2)];

  end
  t=etime(clock,t0);

end
