function [x,hist,t,NoMV]=chebyshev(A,b,x0,mu,radius,tol,kmax)
%x=chebyshev(A,b,x0,mu,radius)
%  x is the computed solution (by Chebyshev Iteration),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector,
%  x0 is the initial guess,
%  mu is the estimated center of the hul of eigenvalues,
%  radius is the estimated radius of this hul.
%
%x=chebyshev(A,b,x0,mu,radius,tol,kmax)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b)
%  kmax maximal number of iterations (100 if not defined)
%
%[x,hist,t,NoMV]=chebyshev(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%

  [n,m]=size(b); 

  %%% set defaults
  if nargin<7, kmax=100; end
  if nargin<6, tol=1.0e-8; end
  if nargin<5, radius=2; end
  if nargin<4, mu=2+1/n; end
  if nargin<3, x0=0*b; end
  
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
     
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end

  t0=clock; 
  %%%
  r0=b-MV(x0);
  rho=norm(r0);    hist=rho;
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  
  nu=mu; r=r0-MV(r0)/mu; x=x0+r0/mu;
  rho=norm(r); hist=[hist,rho];
  radius=radius*radius;
  for k=1:kmax+1
    if rho<tol, break, end
    gamma=radius/nu; nu=2*mu-gamma;
    alpha=2*mu/nu; beta=2/nu; gamma=gamma/nu;
    r2=alpha*r-beta*MV(r)-gamma*r0; r0=r; r=r2;
    x2=alpha*x+beta*r0-gamma*x0;         x0=x; x=x2;
    rho=norm(r); hist=[hist,rho];
  end
  t=etime(clock,t0);

end
