function [x,hist,t,NoMV]=gcrr(A,b,x0,tol,kmax,M,l)
%x=gcrr(A,b)
%  x is the computed solution (by gcr),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=gcrr(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=gcrr(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%
%x=gcrr(A,b,x0,tol,kmax,M)
%  M is the preconditioner, either a matrix, or a function-handle, or 
%    the name of a subroutine, say M='Msolve', that defines 
%    the action of the preconditioner, with u=Msolve(r) we have that u=M\r.
%  If M=[] and u=Msolve(r), then u=r;
%Note that the preconditioner "solves" rather than multiplies.
%
%x=gcrr(A,b,x0,tol,kmax,M,l)
%   l determines restart (l>0) or truncation (l<0).
%   If l>0 then gcr restarts ecvery lth step
%   If l<0 then gcr is trucated with truncation length abs(l)
%   default l=kmax (i.e., gcrr is gcr).
%

  %%% set defaults
  if nargin<3, x0=0*b; end
  if nargin<4, tol=1.0e-8; end
  if nargin<5, kmax=100; end
  if nargin<6 | isempty(M), 
     PRECOND=0;
  else 
     PRECOND=1; 
  end
  if nargin<7, l=kmax; end
 
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
     
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end

  %%% if PRECOND>0, then a preconditioner has been defined
  %%% either as one precondioner or as a product of two factors
  if PRECOND, 
     M_IS_MATRIX=IsMatrix(M);
  end
  
  %%% define the preconditoner MSOLVE
  function u=MSOLVE(r)
    if M_IS_MATRIX, u=M\r; else, u=feval(M,r); end
  end


  t0=clock; 
  %%%
  [n,m]=size(b); 
  c=zeros(n,kmax); u=zeros(n,kmax); sigma=zeros(1,kmax);

  x=x0; r=b; rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
      
  s=1;
  I=repeat(abs(l),kmax);
  
  for k=1:kmax
    if rho<tol, break, end
    kk=I(k);
    if PRECOND,  u(:,kk)=MSOLVE(r); else u(:,kk)=r; end
    c(:,kk)=MV(u(:,kk));
    for j=truncate(l,k):k-1
      jj=I(j);
      beta=(c(:,jj)'*c(:,kk))/sigma(jj);
      u(:,kk)=u(:,kk)-beta*u(:,jj);
      c(:,kk)=c(:,kk)-beta*c(:,jj);
    end
    sigma(kk)=c(:,kk)'*c(:,kk); 
    alpha=(c(:,kk)'*r)/sigma(kk); 
    x=x+alpha*u(:,kk);
    r=r-alpha*c(:,kk);
    rho=norm(r); hist=[hist,rho];
    s=max(k+1-l,1);
  end
  t=etime(clock,t0);

end
%=============================
function s=truncate(l,k)
   s=k;
   if l>0
     s=max(k-l,1);
   elseif l<0
     l=-l;
     s=l*ceil(k/l);
   end

end
%=============================
function I=repeat(ll,kmax)
  I=(1:ll)'*ones(1,ceil(kmax/ll));
  [ll,p]=size(I); I=reshape(I,ll*p,1);
  I=I(1:kmax);
end

