function [x,hist,t,NoMV]=gmres(A,b,x0,tol,kmax,flag)
%x=gmres(A,b)
%  x is the computed solution (by GMRES),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=gmres(A,b,x0,tol,kmax,flag)
%  x0 initial guess (default x0=0*b)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b)
%  kmax maximal number of iterations (100 if not defined)
%  with flag='fom', the FOM method is used instead of GMRES
%
%[x,hist,t,NoMV]=gmres(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.


  %%% set defaults
  if nargin<6, flag='gmres'; end
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end

  FOM=strcmp(lower(flag),'fom');
   
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
     
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
  kappa=-1;
  function [V,H]=ArnStep(V,H)
     w=MV(V(:,k));
     [v,h]=Orth(V,w,kappa);
     V=[V,v];
     H=[[H;zeros(1,k-1)],h];
  end

  t0=clock; 
  %%%
  [n,m]=size(b); 

  x=x0; r=b; rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
     hist=rho; rho0=rho;
  
  V=[r/rho]; H=zeros(1,0); gamma=1; nu=1;
  for k=1:kmax
     if rho<tol, break, end
     [V,H]=ArnStep(V,H);
     %%% compute the nul vector of H'
     gk=(gamma*H(1:k,k))/H(k+1,k); gamma=[gamma,-gk];
     %%% Compute the residual norm
     if FOM
        rho=rho0/abs(gk);
     else
        nu=nu+gk'*gk; rho=rho0/sqrt(nu);
     end
     hist=[hist,rho];
  end
  %%% solve the low dimensional system
  k=size(H,1);
  e=zeros(k-1,1); e(1,1)=1;
  if FOM
     y=H(1:k-1,:)\e;
  else
     e=[e;0]; y=H\e;
  end
  %%% compute the approximate solution 
  x=x0+V(:,1:k-1)*(rho0*y);

  t=etime(clock,t0);

end