function [x,hist,t]=graig(MV,b,x0,tol,kmax,At)
%x=graig(A,b)
%  x is the computed solution (by Graig's method),
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  if A is a matrix, then c=A'*u is used to multiply with the conjugate of A,
%  if A is function, then c=MV(u,'t') is used to multiply with the conjugate of A,
%  b is the right-hand side vector.
%
%x=graig(A,b,x0,tol,kmax)
%  x0 initial guess (0*b if not defined)
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b).
%  kmax maximal number of iterations (100 if not defined).
%
%[x,hist,t,NoMV]=graig(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%


  %%% set defaults
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end

%--------------------------------------------------------------
  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
  
  %%% define the matrix-vector multiplication MV
  function c=MV(u)
    if A_IS_MATRIX, c=A*u; else, c=feval(A,u); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
  TRANSPOSE=max(nargin-5,0);
  %%% TRANSPOSE > 1 if the transpose of A has explicitly been given
  if TRANSPOSE, %%% is the transpose a matrix?
     At_IS_MATRIX=IsMatrix(At);
  end
  
  function ct=MVt(ut)
   if TRANSPOSE
     if At_IS_MATRIX, ct=At*ut; else, ct=feval(At,ut); end 
   else %%% if the transpose of A is not given , use A
     if A_IS_MATRIX, ct=A'*ut; else, ct=feval(A,ut,'t'); end
   end
  end
%-------------------------------------------------------------

  t0=clock; 
  %%%
  [n,m]=size(b); 
  u=zeros(n,1); rho1=1;

  x=x0; r=b(:,1); rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
      hist=rho;
      
  for k=1:kmax
    if rho<tol, break, end
    rho0=rho1; rho1=r'*c; beta=-rho1/rho0;
    c=MVt(r);
    u=c-beta*u;
    c=MV(u);
    sigma=u'*u; alpha=rho1/sigma;
    x=x+alpha*u;
    r=r-alpha*c;
    %%% to save another dot product,
    %%% replace rho by rho1
    rho=norm(r); hist=[hist,rho];
  end
  t=etime(clock,t0);

end
