function [x,hist,t,NoMV]=polynomialsolver(A,b,x0,tol,kmax,Mu,M1,M2)
%x=polynomialsolver(A,b)
%  x is the computed solution (by polynomial resdiual reduction; default LMR).
%  A is a matrix A, or a function-handle, say, A=@MV, 
%    or a function-name, say A='MV', that defines the action of A on a vector: 
%    with c=MV(u) we have that c=A*u;
%  b is the right-hand side vector.
%
%x=polynomialsolver(A,b,x0,tol,kmax,Alpha,M)
%   x0 initial guess (0 if not defined),
%  tol is the required reduction of the residual norm: 
%     at termination norm(b-A*x) < tol*norm(b),
%  kmax maximal number of iterations (100 if not defined).
%  Mu defines the zeros of the error reduction polynomial:
%     with ell=length(Mu), the error reduction polynomial is a power of
%    (1-Mu(1)\lambda)*(1-Mu(2)\lambda)*...*(1-Mu(ell)\lambda).
% In particular of ell=1, we have Richardson iteration with alpha=1/Mu(1).
% If Mu =[] is not defined, 
%    then, per step, the error reduction polynomial is multplied by 
%     (1-alpha*lambda)
%    where, per step, alpha is selected to (locally) minimize residuals (LMR)
%
%[x,hist,t,NoMV]=polynomialsolver(A,b)
%  hist is the sequence norm of intermediate residuals: hist(j)
%    is the 2-norm of the residual after j MVs.
%  t is the computing time, NoMV is the total number of MVs at termination.
%

 
  %%% set default values
  if nargin<6, Mu=[];  end
  if nargin<5, kmax=100; end
  if nargin<4, tol=1.0e-8; end
  if nargin<3, x0=0*b; end
  
  ell=length(Mu);
  if ell, Alpha=1./Mu; end

  %%% Is A a matrix?
  A_IS_MATRIX = IsMatrix(A); NoMV=0;
  %%% if A is not a matrix, then A is a function-handle or a function-name
     
  %%% define the matrix-vector multiplication MV
  function c=MV(x)
    if A_IS_MATRIX, c=A*x; else, c=feval(A,x); end
    %%% count the number of matrix vector multiplications
    NoMV=NoMV+1;
  end
  
  PRECOND=max(nargin-6,0);
  %%% if PRECOND>0, then a preconditioner has been defined
  %%% either as one precondioner or as a product of two factors

  if PRECOND, 
     M1_IS_MATRIX=IsMatrix(M1);
     if PRECOND>1, M2_IS_MATRIX=IsMatrix(M2); end
  end
  
  %%% define the preconditoner MSOLVE
  function u=MSOLVE(r)
    if M1_IS_MATRIX, u=M1\r; else, u=feval(M1,r); end
    if PRECOND==1, return, end
    if M2_IS_MATRIX, u=M2\u; else, u=feval(M2,u); end
  end

  t0=clock;
  %%% j tracks Alpha cycle
  j=1;
 x=x0; r=b; rho=norm(r);
  %%% scale tol for relative residual reduction
  tol=rho*tol;
  %%% Update r0=r if x0 is non-zero
  if norm(x,1)>0, r=r-MV(x); rho=norm(r); end
     hist=rho;

  for k=0:kmax
    if rho<tol, break, end
    if PRECOND, u=Msolve(r); else u=r; end
    c=MV(u);
    if (ell==0)
      %%% minimal residual choice
      alpha=(c'*r)/(c'*c); 
    else
      %%% static polynomial choice
      alpha=Alpha(j); j=j+1; if j>ell, j=1; end
    end
    x=x+alpha*u;
    r=r-alpha*c;
    rho=norm(r); hist=[hist,rho];
  end
  t=etime(clock,t0);

end
