function [v,h]=Orth(V,w,kappa)
%  w=[V,v]*h; with v such that v'*V=0 and v'*v=1;
%
%  kappa=0: classical Gram-Schmidt  (default)
%  kappa>0: repeated Gram-Schmidt with DGKS stopping criterion
%           repeat if tan(angle)<kappa
%  kappa<0: modified Gram-Schmidt
%

%  kappa='hous': Householder reflections

global REPEATED

   if nargin==2, kappa=0; end
   [n,k]=size(V);
   if (k>=n), kappa=0; end
   if (k==0), h=norm(w,2); v=w/h; REPEATED=[]; return, end
   if ~exist('REPEATED'), REPEATED=[]; end


   if ~isnumeric(kappa)
     [v,h]=HouseholderOrth(V,w); 
     return
   end

   v=w;
   if kappa<0 %%% modified Gram-Schmidt

      for j=1:k
        h(j,1)=V(:,j)'*v;
        v=v-V(:,j)*h(j,1);
      end
      rho=norm(v,2); mu=norm(h,2); 

      normw=sqrt(rho*rho+mu*mu);  %%% estimate norm(w)
      zero=k*sqrt(k)*n*normw*eps; %%% numerical zero 

   else %%% repeated Gram-Schmidt
      h=V'*v;
      v=v-V*h;
      rho=norm(v,2); mu=norm(h,2);      t=0;
 
      normw=sqrt(rho*rho+mu*mu);  %%% estimate norm(w) 
      zero=k*sqrt(k)*n*normw*eps; %%% numerical zero     
 
      %%% Daniel Gragg Kaufman Stewart update criterion
      %%% if kappa==0, classical Gram-Schmidt
      while (rho<kappa*mu & rho>zero)
         g=V'*v; v=v-V*g;
         rho=norm(v,2); mu=norm(g,2);
         h=h+g;           
      end
                             REPEATED(1,size(V,2))=t; 
   end

   %%% normalize the expansion vector
   if rho<2*zero %%% w is numerically in span(V)
      %%% if full dimension then no expansion
      if (k>=n), v=zeros(n,0); return, end
      %%% else expand with a random vector
      v=Orth(V,rand(n,1),kappa); 
      h=[h;0]; 
   else
      h=[h;rho];
      v=v/rho;
   end
   
return


%================================================
function [q,h]=HouseholderOrth(Q,w)

global HOUSEHOLDERV HOUSEHOLDERBETA


  [n,j]=size(Q); 
  %%% compute the vectors v for the Householder reflections
  if j==0
    [v,beta,tau]=Householder(w);
    HOUSEHOLDERV=v; HOUSEHOLDERBETA=beta;
    h=tau; q=w/tau;
    return
  end

  for i=1:j
    v=HOUSEHOLDERV(:,i); beta=HOUSEHOLDERBETA(1,i);
    alpha=beta*(v'*w);
    w=w-v*alpha;
  end
  h(1:j,1)=w(1:j,1); j=j+1;
  I=j:n;
  [v,beta,tau]=Householder(w(I,1));
  HOUSEHOLDERV(I,j)=v; HOUSEHOLDERBETA(1,j)=beta; 
  h(j,1)=tau; 
     
  q=zeros(n,1); q(j)=1;
  for i=j:-1:1
    v=HOUSEHOLDERV(:,i); beta=HOUSEHOLDERBETA(1,i);
    alpha=beta*(v'*q);
    q=q-v*alpha;
  end

return
%================================================
function [v,beta,tau]=Householder(a)
%
% [v,beta,tau]=Householder(a)
%     v such that I-beta*v*(v'*a)=tau*e1,
%     where beta=2/(v'*v) and e1=(1;0;0;...;0)
%     v is the vector for the Householder reflection
% 

      v=a;
      tau=norm(v,2);
      g=v(1,1);
      if g~=0, 
         sigma=g/abs(g);  v(1,1)=g+sigma*tau; 
      else 
         sigma=1; v(1,1)=tau; 
      end
      % beta=2/(v'*v); 
      beta=1/(tau*(tau+abs(g)));
      tau=-sigma*tau;
      %% a-v*((v'*a)*beta)
     
return

